#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Schema definitions and resource URIs for the classes implemented by the DRAC
WS-Man API.
"""

DCIM_BIOSEnumeration = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                        'DCIM_BIOSEnumeration')

DCIM_BIOSInteger = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                    'DCIM_BIOSInteger')

DCIM_BIOSService = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                    'DCIM_BIOSService')

DCIM_BIOSString = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                   'DCIM_BIOSString')

DCIM_BootConfigSetting = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                          'DCIM_BootConfigSetting')

DCIM_BootSourceSetting = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                          'DCIM_BootSourceSetting')

DCIM_ComputerSystem = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2'
                       '/DCIM_ComputerSystem')

DCIM_ControllerView = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                       'DCIM_ControllerView')

DCIM_CPUView = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                'DCIM_CPUView')

DCIM_iDRACCardEnumeration = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/'
                             '2/DCIM_iDRACCardEnumeration')

DCIM_iDRACCardInteger = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                         'DCIM_iDRACCardInteger')

DCIM_iDRACCardService = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                         'DCIM_iDRACCardService')

DCIM_iDRACCardString = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                        'DCIM_iDRACCardString')

DCIM_JobService = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                   'DCIM_JobService')

DCIM_LCEnumeration = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                      'DCIM_LCEnumeration')

DCIM_LCService = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                  'DCIM_LCService')

DCIM_LCString = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                 'DCIM_LCString')

DCIM_LifecycleJob = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                     'DCIM_LifecycleJob')

DCIM_MemoryView = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                   'DCIM_MemoryView')

DCIM_NICEnumeration = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                       'DCIM_NICEnumeration')

DCIM_NICInteger = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                   'DCIM_NICInteger')

DCIM_NICService = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                   'DCIM_NICService')

DCIM_NICString = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                  'DCIM_NICString')

DCIM_NICView = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                'DCIM_NICView')

DCIM_PCIeSSDView = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                    'DCIM_PCIeSSDView')

DCIM_PhysicalDiskView = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                         'DCIM_PhysicalDiskView')

DCIM_RAIDEnumeration = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                        'DCIM_RAIDEnumeration')

DCIM_RAIDInteger = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                    'DCIM_RAIDInteger')

DCIM_RAIDService = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                    'DCIM_RAIDService')


DCIM_RAIDString = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                   'DCIM_RAIDString')

DCIM_SystemEnumeration = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                          'DCIM_SystemEnumeration')

DCIM_SystemInteger = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                      'DCIM_SystemInteger')

DCIM_SystemString = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                     'DCIM_SystemString')

DCIM_SystemView = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                   'DCIM_SystemView')

DCIM_VideoView = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                  'DCIM_VideoView')

DCIM_VirtualDiskView = ('http://schemas.dell.com/wbem/wscim/1/cim-schema/2/'
                        'DCIM_VirtualDiskView')
